#!/bin/bash

function lower()
{
    awk ' BEGIN { num_vars = 0 }

    $1 ~ /^bool:/ { var_name[num_vars] = $2; var_idx[$2] = num_vars; ++ num_vars; }

    $1 ~ /^active_var\[0\]=/ { print "**** New constraint ****"; delete av2v; num_active_vars = 0 }

    $1 ~ /^active_var/ { 
        sub(/active_var\[/,"");
        sub(/\]/,"");
        split($0,token,/=/);
        printf("T1=<%s> T2=<%s>\n", token[1], token[2]);
        av2v[num_active_vars] = var_idx[token[2]];
        ++ num_active_vars;
    }

    $1 ~ /^weight/ {
        print $0;
        sub(/weight\["/,"");
        sub(/"\]/,"");
        split($0,token,/=/);
        term_indices = "";
        for (i=0; i<num_active_vars; ++i) {
            c = substr(token[1], i+1, 1);
            if (c ~ /1/) term_indices = term_indices ":" av2v[num_active_vars-i-1];
        }
        sub(/^:/, "", term_indices);
        printf("    term_indices=<%s>\n", term_indices);
        term[term_indices] = term[term_indices] + token[2];
    }

    END {
        print "**** End block ****";

        for (i=0; i<num_vars; ++i) printf("var_name[%d]=%s\n", i, var_name[i]);

        for (vn in var_idx) printf("var_idx[%s]=%d\n", vn, var_idx[vn]);

        for (idx in term) if (term[idx] != 0) printf("term[%s]=%d\n", idx, term[idx]);
    }
'
}

function hubo_to_qubo()
{
    awk '

BEGIN {

    max_orig_ind = 0;
    max_new_ind = 0;

    while (getline < "'"$path/HUBO.2.QUBO"'")
    {
        sub(/\r/,"");

        if ($0 ~ /ORIGTERM/)
        {
            split($0,fields,":");
            sub(/ORIGTERM\[/,"",fields[1]);
            sub(/\]/,"",fields[1]);
            gsub(/,/,":",fields[2]);
            origterm[fields[1]]=fields[2];
            max_orig_ind = 1 + max_orig_ind;
        }

        if ($0 ~ /NEWTERM/)
        {
            split($0,fields,":");
            sub(/NEWTERM\[/,"",fields[1]);
            sub(/\]/,"",fields[1]);
            gsub(/,/,":",fields[2]);
            newterm[fields[1]]=fields[2];
            max_new_ind = 1 + max_new_ind;
        }
    }

    for (i=1; i<=max_orig_ind; ++i)
        lkp[origterm[i]]=newterm[i];
}

{
    split($0, token, "=");
    sub(/term/,"",token[1]);
    printf("term%s=%s\n", lkp[token[1]], token[2]);
}
'
}

function emit_qubo()
{
    awk -v param=${1} '

BEGIN { 
    num_vars = 0;
    if (length(param) > 0) param = " * " param;
}

    $1 ~ /^bool:/ { 
        var_name[num_vars] = $2;
        ++ num_vars;
    }

    $1 ~ /^term/ { 
	split($0, word, /=/);

        sub(/term\[/,"",word[1]);
        sub(/\]/,"",word[1]);
	elements = split(word[1], token, /:/);

	if (elements == 0)
            printf("term: %s%s\n", word[2], param);
	if (elements == 1)
            printf("term: %s%s * %s\n", word[2], param, var_name[token[1]]);
	else if (elements == 2)
            printf("term: %s%s * %s * %s\n", word[2], param, var_name[token[1]], var_name[token[2]]);
    }
'
}

function emit_prod_vars()
{
    awk '

    BEGIN {
        varnum = 0;
        while (getline < "'"$path/varext"'")
        {
            varname[varnum]=$2;
            varnum = varnum + 1;
        }
    }

    $1 ~ /^NEWVAR/ { 
        sub(/^.*:/,"");
        gsub(/\[/,"");
        gsub(/\]/,"");
        split($0,token,"=");
        split(token[2],piece,",");
        printf("# Terms for product variable %s:\n", varname[token[1]]);
        printf("term:  3 * param_prod * %s\n", varname[token[1]]);
        printf("term:  1 * param_prod * %s * %s\n", varname[piece[1]], varname[piece[2]]);
        printf("term: -2 * param_prod * %s * %s\n", varname[piece[1]], varname[token[1]]);
        printf("term: -2 * param_prod * %s * %s\n", varname[piece[2]], varname[token[1]]);
        # print "product=" token[1] " piece1=" piece[1] " piece2=" piece[2];
    }
'
}

function emit_prod_asserts()
{
    awk -v assert_num=$1 '

    BEGIN {
        varnum = 0;
        while (getline < "'"$path/varext"'")
        {
            varname[varnum]=$2;
            varnum = varnum + 1;
        }
    }

    $1 ~ /^NEWVAR/ { 
        sub(/^.*:/,"");
        gsub(/\[/,"");
        gsub(/\]/,"");
        split($0,token,"=");
        split(token[2],piece,",");
        printf("# Assert for product variable %s:\n", varname[token[1]]);
        printf("assert:param_prod: %s * %s - %s\n", varname[piece[1]], varname[piece[2]], varname[token[1]]);
        emit_state_file(varname[token[1]], varname[piece[1]], varname[piece[2]]);
        record_prod_assert();
    }

function emit_state_file(p, a, b,       state_file)
{
    assert_num = 1 + assert_num;
    state_file = sprintf("%s/A%04d", "'"$path"'", assert_num);
    printf("\nparam: param_prod\n") >> state_file;
    printf("\nassert: '"'"'%s - %s * %s'"'"'\n\n", p, a, b) >> state_file;
    printf("active_vars[0]=%s\n", b) >> state_file;
    printf("active_vars[1]=%s\n", a) >> state_file;
    printf("active_vars[2]=%s\n\n", p) >> state_file;
    printf("%s %s %s value\n", p, a, b) >> state_file;
    printf("-------- -- -- --\n") >> state_file;
    printf("       0  0  0  1\n") >> state_file;
    printf("       0  0  1  1\n") >> state_file;
    printf("       0  1  0  1\n") >> state_file;
    printf("       0  1  1  0\n") >> state_file;
    printf("       1  0  0  0\n") >> state_file;
    printf("       1  0  1  0\n") >> state_file;
    printf("       1  1  0  0\n") >> state_file;
    printf("       1  1  1  1\n") >> state_file;
}

function record_prod_assert()
{
    prod_list_file = sprintf("%s/%s", "'"$path"'", "prodlist");
    printf("A%04d\n", assert_num) >> prod_list_file;
}
'
}

path=$(dw get path)
path=${path%\'}
path=${path##*\'}

# STEP 0:

rm -f \
    $path/A*.posiform.HUBO \
    $path/A*.posiform.QUBO \
    $path/HUBO.2.QUBO \

# STEP 1:

for f in $path/A*.posiform
do
    cat $path/variable $f | lower | grep 'term\[' > ${f}.HUBO
done

# STEP 2:

os=$( uname | tr '[:upper:]' '[:lower:]' )

if [[ $os == ming* ]]
then
    sort=/c/MinGW/msys/1.0/bin/sort.exe
else
    sort=sort
fi

awk '{sub(/=.*/,"=0"); print $0}' $path/A*.posiform.HUBO |

$sort | uniq |

${DWAVE_HOME}/bin/dw-lower > $path/HUBO.2.QUBO

# STEP 3:

for f in $path/A*.posiform.HUBO
do
    hubo_to_qubo <${f} >${f%HUBO}QUBO
done

{
    cat $path/variable
    awk ' /^NEWVAR/ { sub(/\]=.*$/,"");
                      sub(/^.*:\[/,"");
                      print "bool: prodvar" $0}' <$path/HUBO.2.QUBO
} >$path/varext

# STEP 4:

[[ -e $path/parameter ]] && cat $path/parameter

grep -q NEWVAR $path/HUBO.2.QUBO && echo "param: param_prod"

awk '{sub(/^bool:/,"var:"); print $0}' $path/varext

emit_prod_vars <$path/HUBO.2.QUBO

max_assert_num=-1

shopt -s extglob

for f in $path/A*.posiform.QUBO
do
    stem=${f##*/}
    assertion=${stem%.posiform.QUBO}
    echo "# Assertion: ${assertion}"
    param=$(awk '/^param:/ {print $2}' ${path}/${assertion})
    cat $path/varext ${f} | emit_qubo ${param}

    assertion=$(awk '/^assert:/ \
                    {sub(/^assert: /,""); gsub(/'"'"'/,""); print $0;}' \
                ${path}/${assertion})

    if [[ -n ${param} ]]
    then
	echo "assert:${param}: ${assertion}"
    else
	echo "assert: ${assertion}"
    fi

    this_assert_num=${f##${path}/A*(0)}
    this_assert_num=${this_assert_num%.posiform.QUBO}
    (( this_assert_num > max_assert_num )) && max_assert_num=$this_assert_num
done

emit_prod_asserts $max_assert_num <$path/HUBO.2.QUBO
