#!/bin/bash

# This shell script uses a set of variables and assertions as input.
# From there it performs a recursive depth-first tree search to find
# simultaneous solutions to the entire set of assertions.  Note that
# the assertions are specified via "dw set assert ..." and this program
# reads the state tables which are produced as a side effect.  Thus,
# this program is completely independent of the HUBO/QUBO representations
# that may be generated from the assertions.

{
    dw get variable

    dw get assert |
    while read assert_name etc
    do
        echo "**** Assertion ${assert_name%:} ****"
	dw get assert ${assert_name%:}
    done
} |

awk '

BEGIN { SUBSEP = ","; vars = 0; asserts = 0; }

$1 ~ /^bool:$/ { var_index[$2] = vars; var_name[vars] = $2; ++ vars }

$2 ~ /Assertion/ {
    ++ asserts;
    sub(/A/,"",$3);
    anum = 0 + $3;
    num_vars[anum] = 0;
    num_states[anum] = 0;
    s = 0;
}

$1 ~ /^active_vars.*/ {
    sub(/^.*=/,"");
    V[anum,num_vars[anum]] = var_index[$0];
    ++ num_vars[anum];
    states = 2 ** num_vars[anum];
    value_column = 1 + num_vars[anum];
}

$1 ~ /^0$|^1$/ {
    A[anum,num_states[anum]] = $value_column;
    ++ num_states[anum];
}

END {
    for (i=0; i<vars; ++i) var_value[i] = -1;
    found_solution = 0;
    search(0);
    exit 1-found_solution;
}

function search(level,      i, ic, j, k, bit, ok, unset_me)
{
    if (level == asserts)
    {
        for (i=0; i<vars; ++i) printf("%s=%d ", var_name[i], var_value[i]);
        printf("\n");
        found_solution = 1;
        return;
    }

    for (i=0; i<num_states[level]; ++i)
    {
        if (A[level,i] == 0) continue;

        ic = i;

        ok = 1;
        k = 0;
	for (j=0; j<num_vars[level]; ++j)
	{
            bit = ic % 2;
            ic = (ic - bit) / 2;

            if (var_value[V[level,j]] == -1)
            {
                var_value[V[level,j]] = bit;
                unset_me[k] = V[level,j];
                ++ k;
            }
            else if (var_value[V[level,j]] != bit)
            {
                ok = 0;
                break;
            }
	}

        if (ok) search(level+1);

	for (j=0; j<k; ++j) var_value[unset_me[j]] = -1;
    }
}

'
