#!/bin/bash

# Build a QUBO which has the structure of a 4-cycle.
# The QUBO makes one of each pair of adjacent variables turn on.
# Embed the QUBO (which should require exactly 4 qubits)
# and run it.  Two ground states should emerge.

# Exit if not being run via Bash shell:
if [ ! -n "$BASH" ] ; then 
    echo Please run this script $0 with bash; exit 1; 
fi

# Get a connection to the local simulator:
dw set connection laptop

# Get a 128-qubit sampling solver from the simulator:
dw set solver c4-sw_sample

# Make new workspace directory (examples) if it does not exist:
dw mkdir -p examples

# Change to new workspace directory:
dw cd examples

# Initialize the workspace:
dw init

# Create the QUBO:
echo "
var: w
var: x
var: y
var: z

term: -1 * w
term: -1 * x
term: -1 * y
term: -1 * z

term: w * x
term: x * y
term: y * z
term: z * w

assert: w + x - 1
assert: x + y - 1
assert: y + z - 1
assert: z + w - 1
" > D-qubo.q

# Embed the QUBO:
dw embed D-qubo.q

# There are no parameters, so issue an "empty" bind command to create the QMI:
dw bind empty

# Execute the QMI:
dw exec num_reads=1000 default.qmi

# Display the first solution:
dw val -v -s 1 default.sol

# Display the second solution:
dw val -v -s 2 default.sol
