#!/bin/bash

# Illustrates the use of user-defined macros with the dw constraint compiler.
# Create a user-defined macro and use it in an assertion, along with one of
# built-in assertions.  Generate a QUBO, embed it and run.

# Exit if not being run via Bash shell:
if [ ! -n "$BASH" ] ; then 
    echo Please run this script $0 with bash; exit 1; 
fi

# Get a connection to the local simulator:
dw set connection laptop

# Get a 128-qubit sampling solver from the simulator:
dw set solver c4-sw_sample

# Make new workspace directory (examples) if it does not exist:
dw mkdir -p examples

# Change to new workspace directory:
dw cd examples

# Initialize workspace:
dw init

# Create a user-defined macro:
dw set macro balance '($1 + $2) - ($3 + $4)'

# Define variables:
dw set variable w
dw set variable x
dw set variable y
dw set variable z
dw set variable o

# Define assertions:
dw set assert 'balance(w,x,y,z)'
dw set assert 'R2_ne(z,o)'

# Compile the constraints to create the QUBO:
dw compile > H.q

# Embed the QUBO:
dw embed H.q

# Set the chain strength parameter via bind:
dw bind param_chain=5

# Execute the QMI:
dw exec num_reads=1000 default.qmi

# Display the sample statistics:
dw val -v default.sol

# Display the first six solutions:
for ((s=1; s<=6; ++s))
do
    dw val -v -s $s default.sol
done
