#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>

#include "qsage.h"

unsigned int *RTR_edge_color;
unsigned int  RTR_vertices;
unsigned int  RTR_clique_size;
unsigned int  RTR_clique[32];
unsigned int  RTR_mask;
unsigned int  RTR_valid_color_shifted;

#define edge_color(r,c) RTR_edge_color[(r)*RTR_vertices+(c)]

unsigned int ramsey_tree_recursive(unsigned int level,
				   unsigned int next_vertex);

int ramsey_init()
{
  printf("RAM_red = %d\n", RAM_red);
  printf("RAM_blue = %d\n", RAM_blue);
  printf("RAM_vertices = %d\n", RAM_vertices);

  RTR_vertices = RAM_vertices;
  RTR_mask = 1;

  RTR_edge_color = malloc(RTR_vertices * RTR_vertices * sizeof(unsigned int));
  memset(RTR_edge_color, 0, RTR_vertices * RTR_vertices * sizeof(unsigned int));

  return RTR_vertices * (RTR_vertices - 1) / 2;
}

double ramsey_obj(const int* state, size_t len)
{
  double obj = 0.0;
  int r, c, j = 0;

  for (r=0; r<RTR_vertices; ++r)
    for (c=r+1; c<RTR_vertices; ++c)
      edge_color(r,c) = state[j++];

  RTR_valid_color_shifted = 0;
  RTR_clique_size = RAM_red;
  obj += ramsey_tree_recursive(0, 0);

  RTR_valid_color_shifted = 1;
  RTR_clique_size = RAM_blue;
  obj += ramsey_tree_recursive(0, 0);

  return obj;
}

unsigned int ramsey_tree_recursive(unsigned int level,
				   unsigned int next_vertex)
{
  unsigned int v_curr, v_prior;
  unsigned int clique_member;
  unsigned int is_valid;
  unsigned int edge_color;
  unsigned int objective = 0;

  for (v_curr=next_vertex; v_curr<RTR_vertices; ++v_curr)
    {
      is_valid = 1;

      for (clique_member=0; is_valid && clique_member<level; ++clique_member)
	{
	  v_prior = RTR_clique[clique_member];
	  edge_color = RTR_edge_color[v_prior*RTR_vertices + v_curr];
	  is_valid &= (edge_color & RTR_mask) == RTR_valid_color_shifted;
	}

      if (level == RTR_clique_size-1)
	objective += is_valid;
      else if (is_valid)
	{
	  RTR_clique[level] = v_curr;
	  objective += ramsey_tree_recursive(level+1, v_curr+1);
	}
    }

  return objective;
}

char *red = "red";
char *blue = "blue";

void ramsey_final(const int* state, size_t len)
{
  int r, c, j = 0;
  char *color;

  for (r=0; r<RTR_vertices; ++r)
    for (c=r+1; c<RTR_vertices; ++c)
      {
	edge_color(r,c) = state[j++];
	if (edge_color(r,c) == 0) color = red;
	if (edge_color(r,c) == 1) color = blue;
	printf("color(%d,%d) = %s\n", r,c, color);
      }
}
